#

/* This file contains the CHAOS packet forwarders (which run in
 * the GW process) as well as conversion routines. These routines
 * do forwarding for old style CHAOS packets as well as the
 * conversion between old and new formats. At some point they can
 * be thrown away.
 *
 * NOTE: Neither of the conversion routines has been tested and they
 * contain known bugs!
 */


#include	<defs.h>
#include	"../../mos/mos.h"
#include	"../../aux/mosu.h"
#include	"../src/macs.h"
#include	"../src/const.h"
#include	"../src/param.h"
#include	"../src/net.h"
#include	"../src/ext.h"
#include	"../in/in.h"
#include	"../in/inext.h"
#include	"ch.h"
#include	"chparam.h"
#include	"chext.h"


/* Error messages */

char chpdsc[]	"Packet %o-%o from %o/%o-%o to %o/%o-%o discarded rsn %o\n";



/* Old style header processing; check to make sure the packet doesn't
 * need conversion; if so, do it and use the regular IN code to do
 * the forwarding, otherwise do the forwarding. Note that the reasonability
 * checks are performed even for internally generated packets.
 */

ochfwd(iob)
iorb	*iob;

{	reg	ochpkt	*pkt;
	reg	net	*netp;
	reg	unsb	snet;
		chaddr	*hst;
		chia	*iap;
		word	code;
		inpkt	*onchcnv();
		word	netsend();


	pkt = mkoch(iob->i_addr);

	if (pkt->oc_pad != 0)
		bughalt("not CHAOS packet");
	if (++pkt->oc_fc == 0)
		bughalt("fc ran out");
	if ((pkt->oc_len + sizeof(ochpkt)) > iob->i_bxfr)
		bughalt("truncated CHAOS packet");

	if (pkt->oc_dst.oc_word == 0) {
		signal(chpid, S_PKT, iob);
		return;
		}

	if ((snet = pkt->oc_dst.oc_snt) == 0)
		bughalt("CHAOS pkt for subnet 0");
	if (snet >= CHMSNET)
		bughalt("table ovflo");
	if (ochtab[snet] != 0) {
		iob->i_addr = ((char *) onchcnv(pkt));
		ninfwd(iob);
		return;
		}

	if ((netp = chitab[snet]) == NULL)
		bughalt("cgtfh");
	if (chttab[snet] != T_DIR)
		hst = &chgatab[snet];
	  else {
		hst = &pkt->oc_dst;

		iap = chiatlst[netp->n_net];
		while (iap != NULL) {
			if (pkt->oc_dst.oc_word ==
					mkcha(&iap->chia_addr)->oc_word) {
				signal(chpid, S_PKT, iob);
				return;
				}
			iap = iap->chia_link;
			}
		}

	iob->i_breq = (pkt->oc_len + sizeof(ochpkt));
	if (iob->i_breq > netp->n_max)
		bughalt("CHAOS pkt too big");

	if (netp->n_send != NULL)
		code = (*netp->n_send)(iob, netp, P_CH, hst);
	  else {
		(*netp->n_out)(iob, netp, P_CH, hst);
		code = netsend(iob, netp);
		}
	if (code != D_OK) {
		log(L_UNUS, &chpdsc, pkt->oc_pkt, pkt->oc_ack,
				mkscha(&pkt->oc_src), pkt->oc_sind,
				mkscha(&pkt->oc_dst), pkt->oc_dind, code);
		freebuf(bufsiz, iob);
		}
}


/* An IN based CHAOS packet has been received. See if it is
 * for a subnet that use old style format; if so, convert it
 * and call the old style forwarder, otherwise give it to the
 * regular IN forwarder.
 * The reasonability checks are performed here before conversion
 * since this is called from before the IN forwarder performed these
 * checks.
 * Note the slight duplication of effort, since the old style
 * forwarder will check to make sure it doesn't need conversion.
 */

nchfwd(iob)
reg	iorb	*iob;

{	reg	inpkt	*pkt;
	reg	unsw	tmp;
		ochpkt	*nochcnv();

	pkt = mkin(iob->i_addr);

	if ((pkt->i_dst.i_byteh == myanet) &&
	    (ochtab[pkt->i_dst.i_ssnet] == 0)) {

		bughalt("CHAOS/IN pkt");

		if ((pkt->i_ihlver & INHVERM) != INHVER)
			bughalt("not V4 packet");
		tmp = (pkt->i_ihlver & INHLENM);
		if (tmp < INMINH)
			bughalt("bad header length");
		tmp =* INWPHW;

		if ((cksum(pkt, tmp, 0) + 1) != 0)
			bughalt("header fried");
		if (--pkt->i_ttl == 0)
			bughalt("ttl ran out");

		if (swab(pkt->i_len) > iob->i_bxfr)
			bughalt("truncated IN packet");
		tmp = swab(pkt->i_fofflg);
		if ((tmp & INUNUF) != 0)
			bughalt("unused flag on");
		if (((tmp & INMFRG) != 0) || ((tmp & INFOFFM) != 0))
			bughalt("gack, fragmented packet");

		if (pkt->i_dst.i_srsd != 0)
			bughalt("bad CHAOS address");

		iob->i_addr = ((char *) nochcnv(pkt));
		ochfwd(iob);
		return;
		}

	ninfwd(iob);
}


/* Convert old style to new style. The code is very careful about
 * not stepping on its own toes; don't change the order of the
 * assignment statements without checking to make sure you don't
 * trash anything. Note that the dest host is already in the 
 * right place.
 */

inpkt	*onchcnv(pkt)
reg	ochpkt	*pkt;

{	reg	inpkt	*npkt;
	reg	nchpkt	*nhdr;
	unsb	opcod;


	bughlt("conversion called");

	npkt = (((char *) pkt) - ((sizeof(inpkt) + sizeof(nchpkt))
							 - sizeof(ochpkt)));
	nhdr = (((char *) npkt) + sizeof(inpkt));

	opcod = pkt->oc_opc;

	npkt->i_ihlver = (INHVER | INMINH);
	npkt->i_tos = 0;
	npkt->i_len = swab(pkt->oc_len + (sizeof(inpkt) + sizeof(nchpkt)));
	npkt->i_id = pkt->oc_pkt;
	npkt->i_fofflg = 0;
	npkt->i_ttl = -pkt->oc_fc;	/* Count in different direction */
	npkt->i_prot = INCH;
	npkt->i_src.i_snet = myanet;
	npkt->i_src.i_ssnet = pkt->oc_src.oc_snt;
	npkt->i_src.i_srsd = 0;
	npkt->i_src.i_shst = pkt->oc_src.oc_hst;
	npkt->i_dst.i_snet = myanet;
	npkt->i_dst.i_ssnet = pkt->oc_dst.oc_snt;
	npkt->i_dst.i_srsd = 0;
	npkt->i_cksum = 0;
	npkt->i_cksum = ~cksum(npkt, (INMINH * INWPHW), 0);
	if (npkt->i_cksum == 0)
		npkt->i_cksum = -1;

	nhdr->nc_dind = pkt->oc_dind;
	nhdr->nc_sind = pkt->oc_sind;
	nhdr->nc_pkt = pkt->oc_pkt;
	nhdr->nc_ack = pkt->oc_ack;
	nhdr->nc_opc = opcod;
	nhdr->nc_pad = 0;
	nhdr->nc_cksum = 0;

	return(npkt);
}


/* Convert new style header to old. Can't pull the dest hst hack since there
 * may be options.
 */

ochpkt	*nochcnv(pkt)
reg	inpkt	*pkt;

{	reg	ochpkt	*npkt;
	reg	nchpkt	*hdr;
		unsb	opcod;
		unsw	tmp;


	bughlt("conversion called");

	hdr = (((char *) pkt) + (INBPHW * (pkt->i_ihlver & INHLENM)));
	npkt = (((char *) hdr) - (sizeof(ochpkt) - sizeof(nchpkt)));

/*	if (hdr->nc_cksum != 0) {
		if (((tmp = swab(pkt->i_len)) & 01) != 0)
			((byte *) pkt)[tmp] = 0;
		if (cksum(***, (++tmp >> 1), 0) != 0)
			bughalt("bad checksum");
		}	*/

	opcod = hdr->nc_opc;

	npkt->oc_pkt = hdr->nc_pkt;
	npkt->oc_ack = hdr->nc_ack;
	npkt->oc_sind = hdr->nc_sind;
	npkt->oc_dind = hdr->nc_dind;
	npkt->oc_src.oc_snt = pkt->i_src.i_ssnet;
	npkt->oc_src.oc_hst = pkt->i_src.i_shst;
	npkt->oc_dst.oc_snt = pkt->i_dst.i_ssnet;
	npkt->oc_dst.oc_hst = pkt->i_dst.i_shst;
	npkt->oc_fc = -pkt->i_ttl;
	npkt->oc_len = swab(pkt->i_len);
	npkt->oc_opc = opcod;
	npkt->oc_pad = 0;

	return(npkt);
}
