#

/* This file includes CHAOS startup and input packets demultiplexing
 * as well as support for all CHAOS network support tasks; routing
 * generation and listening, etc. Packets coming in from
 * the nets have their IORB pointers passed to this process.
 */


#include	<defs.h>
#include	"../../mos/mos.h"
#include	"../../aux/mosu.h"
#include	"../src/defs.h"
#include	"../src/macs.h"
#include	"../src/const.h"
#include	"../src/param.h"
#include	"../src/net.h"
#include	"../src/ext.h"
#include	"ch.h"
#include	"chparam.h"
#include	"chext.h"



/* Useful constants */

#define	CHSP	040	/* Space in CHAOS char set */
#define	CHNL	010	/* NL in CHAOS char set */


/* Error and info messages */

char chnopc[]	"No handler for opcode %o from host %o/%o, ignored\n";
char chnsrv[]	"No server for RFC to %s from host %o/%o, ignored\n";
char chcls[]	"Sending CLS of %s to host %o/%o\n";



/* CHAOS protocol startup. Sets up some initial stuff, clears out the
 * routing tables, rebuild the initial routing tables from device
 * device interface info and then calls the per network protocol
 * initialization for networks that we know about. Finally it starts
 * the mux, which accepts all CHAOS packets addressed to this host.
 */

CHNET()

{	ext	word	ochfwd();
	word	chmux();	

	chopc = addshnd(gwhnd, &ochfwd);
	chpid = gapid();
	chhnd = init_hnd(MXCHSIG);
	if (addshnd(chhnd, &chmux) != S_PKT)
		bughalt("handler bad sig");

	clrchtabs();
	initchtabs();
	chprinit();

	run_hnd(chhnd);
}


/* Clear all CHAOS routing tables. Note that the gateway tables are not
 * cleared since you can tell merely by looking at the type field
 * what kind of connection you are dealing with, and all non direct
 * and fixed routes are cleared at startup.
 */

clrchtabs()

{	reg	net	**nip;
	reg	unsw	*gtp;
	reg	unsw	*gcp;
		unsb	snet;

	nip = &chitab;
	gtp = &chttab;
	gcp = &chgctab;
	for (snet = 0; snet < CHMSNET; snet++) {
		*nip++ = NULL;
		*gtp++ = T_NONE;
		*gcp++ = CHMXCST;
		}
}


/* Enter entries from hardware interface address tables.
 */

initchtabs()

{	reg	unsb	snet;

	{	reg	net	*netp;
		reg	chia	*iap;

		for (netp = &nets[0]; netp < lstnet; netp++) {
			if ((iap = chiatlst[netp->n_net]) == NULL)
				continue;
			if (netp->n_prinit != NULL)
				(*netp->n_prinit)(netp, P_CH, iap);

			while (iap != NULL) {
				if ((snet = mkcha(&iap->chia_addr)->oc_snt)
									== 0)
					bughalt("Illegal subnet");
				if (snet >= CHMSNET)
					bughalt("Table overflow");
				if (chttab[snet] != T_NONE)
					bughalt("Table overwrite");
				chitab[snet] = netp;
				chttab[snet] = T_DIR;
				chgctab[snet] = iap->chia_cost;
				iap = iap->chia_link;
				}
			}
	}

	{	reg	chfre	*frep;
		reg	unsb	dsnet;
			unsb	nfre;
		
		frep = &chfretbl;
		for (nfre = 0; nfre < nchfre; nfre++, frep++) {
			if ((snet = frep->chfre_snt) == 0)
				bughalt("Illegal subnet");
			if (snet >= CHMSNET)
				bughalt("Table overflow");
			if (chttab[snet] != T_NONE)
				bughalt("Table overwrite");
			dsnet = mkcha(&frep->chfre_addr)->oc_snt;
			if (chttab[dsnet] != T_DIR)
				bughalt("Bad fixed route");
			chitab[snet] = chitab[dsnet];
			chttab[snet] = T_FIX;
			chgatab[snet].oc_word =
					mkcha(&frep->chfre_addr)->oc_word;
			chgctab[snet] = frep->chfre_cost;
			}
		}
}


/* Top level for CHAOS protocol services; basically waits to
 * be handed packets from the forwarding process. Note that the
 * way packets get sent is that an event handler is added to the
 * GWAY process that calls the CHAOS forwarder directly, as if
 * a packet had been received on an interface! Needs to send a LOS.
 */

word	chmux(iob)
reg	iorb	*iob;

{	reg	ochpkt	*pkt;

	pkt = mkoch(iob->i_addr);

	switch (mkunsb(pkt->oc_opc)) {

	  case CHRUT:	chrut(iob);
			break;

	  case CHRFC:	chrfc(iob);
			break;

	  default:	log(L_ERR, chnopc, mkunsb(pkt->oc_opc),
				mkscha(&pkt->oc_src));

	  case CHMNT:	freebuf(bufsiz, iob);
			break;
	  }
}


/* Process incoming RFC's. Parses the string and converts it to a C
 * style string, and then, if there are not arguments, provides
 * a null string for the 'rest', otherwise converting the rest to a
 * C string.
 */

chrfc(iob)
reg	iorb	*iob;

{	reg	ochpkt	*pkt;
	reg	char	*srvr;
	unsw	len	0;
	unsw	plen;

	pkt = mkoch(iob->i_addr);
	srvr = &pkt->oc_data[0];
	plen = pkt->oc_len;

	while ((len < plen) && (*srvr != CHSP)) {
		srvr++;
		len++;
		}
	*srvr++ = '\0';
	if (len == plen)
		*srvr = '\0';
	  else
		pkt->oc_data[plen] = '\0';

	if (strcmp(&pkt->oc_data, "DUMP-ROUTING-TABLE") == 0) {
		chdrutt(iob);
		return;
		}
	if (strcmp(&pkt->oc_data, "STATUS") == 0) {
		chstatus(iob);
		return;
		}
	if (strcmp(&pkt->oc_data, "PULSAR") == 0) {
		chpulsar(iob, srvr);
		return;
		}

	log(L_ERR, chnsrv, &pkt->oc_data, mkscha(&pkt->oc_src));
	chmkcls(iob, "No server for RFC");
}


/* Routine to turn a packet into an ANS. */

chmkans(pkt)
reg	ochpkt	*pkt;

{	word	src;

	pkt->oc_opc = CHANS;
	pkt->oc_fc = 0;
	src = pkt->oc_dst.oc_word;
	pkt->oc_dst.oc_word = pkt->oc_src.oc_word;
	pkt->oc_src.oc_word = src;
	pkt->oc_dind = pkt->oc_sind;
	pkt->oc_sind = 0;
	pkt->oc_ack = pkt->oc_pkt;
	pkt->oc_pkt = 0;
}


/* Routine to turn a packet into an CLS with a string error message
 * and zap it off.
 */

chmkcls(iob, errstr)
reg	iorb	*iob;
char	*errstr;

{	reg	ochpkt	*pkt;
	reg	char	*p, *q;
	word	src;
	unsw	len	0;

	pkt = mkoch(iob->i_addr);
	log(L_ERR, chcls, errstr, mkscha(&pkt->oc_src));
	pkt->oc_opc = CHCLS;
	pkt->oc_fc = 0;
	src = pkt->oc_dst.oc_word;
	pkt->oc_dst.oc_word = pkt->oc_src.oc_word;
	pkt->oc_src.oc_word = src;
	pkt->oc_dind = pkt->oc_sind;
	pkt->oc_sind = 0;
	pkt->oc_ack = pkt->oc_pkt;
	pkt->oc_pkt = 0;

	p = errstr;
	q = &pkt->oc_data[0];
	while (*p != '\0') {
		*q++ = *p++;
		len++;
		}
	pkt->oc_len = len;
	iob->i_bxfr = (len + sizeof(ochpkt));
	signal(gwpid, chopc, iob);	/* Fake out the checks in the fwder */
}
