#

/* This file includes support for all CHAOS network routing
 * related tasks; table generation and listening, etc.
 */


#include	<defs.h>
#include	"../../mos/mos.h"
#include	"../../aux/mosu.h"
#include	"../src/defs.h"
#include	"../src/macs.h"
#include	"../src/const.h"
#include	"../src/param.h"
#include	"../src/net.h"
#include	"../src/ext.h"
#include	"ch.h"
#include	"chparam.h"
#include	"chext.h"


/* RUT packets consist of a list of these entries */

struct	chrutb	{	unsw	cr_snet;
			unsw	cr_cost;
		};


/* The ANS to a DUMP-ROUTING-TABLE RFC contains lists of these
 * entries.
 */

struct	chdrtb	{	chaddr	cd_addr;
			unsw	cd_cost;
		};


/* Error and info messages */

char chbds[]	"Bad subnet 0 route from host %o/%o ignored\n";
char chovfl[]	"Routing table overflow, subnet %o from host %o/%o, ignored\n";
char chcost[]	"Low cost of %d for subnet %o from host %o/%o ignored\n";
char chnre[]	"Old route of %d to %o/%o for %o replaced, new %d to %o/%o\n";
char chrutd[]	"Route of %d to host %o/%o for subnet %o decayed\n";
char chdrt[]	"Responding to DUMP-ROUTING-TABLE from host %o/%o\n";



/* Process incoming routing packets, an entry at a time, seeing
 * if any useful information is contained.
 */

chrut(iob)
iorb	*iob;

{	reg	ochpkt	*pkt;
	reg	struct	chrutb	*rdat;
		chaddr	*hst;
		unsw	*cst;
	reg	unsb	snet;
		unsb	dsnet;
		unsw	ent;

	pkt = mkoch(iob->i_addr);
	rdat = ((struct chrutb *) &pkt->oc_data);

	for (ent = (pkt->oc_len / sizeof(struct chrutb));
					 ent > 0; ent--, rdat++) {

		if ((snet = rdat->cr_snet) == 0) {
			log(L_ERR, chbds, mkscha(&pkt->oc_src));
			continue;
			}
		if (snet >= CHMSNET) {
			log(L_ERR, chovfl, snet, mkscha(&pkt->oc_src));
			continue;
			}
		cst = &chgctab[snet];
		if (*cst <= rdat->cr_cost)
			continue;
		switch(chttab[snet]) {

		  case T_DIR:
		  case T_FIX:	log(L_ERR, chcost, rdat->cr_cost, snet,
					mkscha(&pkt->oc_src));
				break;

		  case T_NONE:	chttab[snet] = T_FLT;
				chgatab[snet].oc_word = ~0;

		  case T_FLT:	hst = &chgatab[snet];
				*cst = rdat->cr_cost;

				if (hst->oc_word == pkt->oc_src.oc_word)
					break;

				log(L_INF, chnre, *cst, mkscha(hst), snet,
					rdat->cr_cost, mkscha(&pkt->oc_src));

				if ((dsnet = pkt->oc_src.oc_snt) == 0)
					bughalt("Illegal subnet 0 route");
				if (dsnet >= CHMSNET)
					bughalt("Bad source subnet route");
				if (chttab[dsnet] != T_DIR)
					bughalt("bad routing data");

				chitab[snet] = chitab[dsnet];
				hst->oc_word = pkt->oc_src.oc_word;
				break;

		  default:	bughalt("bad network type code");
		  }
  	}

	freebuf(bufsiz, iob);
}


/* Age routing info. Bogus type stuff is C lossage.
 */

word	charut()

{	reg	unsb	*gtp;
	reg	unsw	*gcp;
	reg	unsb	snet;

	gtp = &chttab;
	gcp = &chgctab;
	for (snet = 0; snet < CHMSNET; snet++, gtp++, gcp++) {
		if (*gtp != T_FLT)
			continue;
		if (++(*gcp) >= CHMXCST) {
			log(L_ERR, chrutd, *gcp, snet, mkscha(&chgatab[snet]));
			chitab[snet] = NULL;
			*gtp = T_NONE;
			}
		}
}


/* Routine that actually sends out the routing packets. For CHAOS/IN
 * conversion, the routing packets are only broadcast for virtual
 * subnets that are old style. Loops over the network table and
 * creates a routing packet for each network, which it then queues.
 */

word	chxrut()

{	reg	net	*netp;
	reg	chia	*iap;
	reg	iorb	*iob;
		unsw	snet;
		iorb	*chmkrut();
		iorb	*netsend();

	for (netp = &nets[0]; netp < lstnet; netp++) {
		if ((netp->n_cap & C_BRD) == 0)
			bughalt("not broadcast net");
		for (iap = chiatlst[netp->n_net]; iap != NULL;
						iap = iap->chia_link) {

			snet = mkcha(&iap->chia_addr)->oc_snt;
			if (ochtab[snet] != 0)
				continue;

			iob = chmkrut(chgctab[snet]);
			mkoch(iob->i_addr)->oc_src.oc_word =
					mkcha(&iap->chia_addr)->oc_word;
			if (iob->i_breq > netp->n_max)
				bughalt("CHAOS pkt too big");

			if (netp->n_send != NULL)
				iob = (*netp->n_send)(iob, netp, P_CH, ANYHOST);
			  else {
				(*netp->n_out)(iob, netp, P_CH, ANYHOST);
				iob = netsend(iob, netp);
				}
			if (iob != NULL)
				bughalt("ch rut pkt op ovflo");
			}
		}
}


/* Routine that creates the routing packets. Loops over the routing
 * table and generates a RUT packet for this net. Note code to deliberately
 * skip subnet zero.
 */

iorb	*chmkrut(cst)
unsw	cst;

{	reg	iorb	*iob;
	reg	ochpkt	*pkt;
	reg	struct	chrutb	*rdat;
		unsb	*gtp;
		unsb	snet;
		unsw	nents	0;

	if ((iob = mkiorb(getbuf(bufsiz))) == NULL)
		bughalt("Can't allocate packet for CHAOS routing broadcast");
	pkt = mkoch(((byte *) iob) + pktoff);

	pkt->oc_pad = 0;
	pkt->oc_opc = CHRUT;
	pkt->oc_fc = 0;
	pkt->oc_sind = 0;
	pkt->oc_dst.oc_word = 0;
	pkt->oc_dind = 0;
	pkt->oc_pkt = 0;
	pkt->oc_ack = 0;

	rdat = ((struct chrutb *) &pkt->oc_data);
	gtp = &chttab[1];
	for (snet = 1; snet < CHMSNET; snet++, gtp++) {
		if (*gtp == T_NONE)
			continue;
		rdat->cr_snet = snet;
		rdat->cr_cost = (chgctab[snet] + cst);
		rdat++;
		nents++;
		}

	pkt->oc_len = (sizeof(struct chrutb) * nents);
	iob->i_breq = (pkt->oc_len + sizeof(ochpkt));
	iob->i_addr = ((byte *) pkt);
	return(iob);
}


/* Process DRT RFC's. Simply dump out the outing table.
 */

chdrutt(iob)
iorb	*iob;

{	reg	ochpkt	*pkt;
	reg	struct	chdrtb	*rtdbp;
	reg	unsw	snet;

	pkt = mkoch(iob->i_addr);
	log(L_INF, &chdrt, mkscha(&pkt->oc_src));

	chmkans(pkt);
	rtdbp = ((struct chdrtb *) &pkt->oc_data);

	for (snet = 0; snet < CHMSNET; snet++, rtdbp++) {
		switch(chttab[snet]) {

		  case T_NONE:	rtdbp->cd_addr.oc_word = 0;
				rtdbp->cd_cost = CHMXCST;
				break;

		  case T_DIR:	rtdbp->cd_addr.oc_word = (1 + (2 *
					chitab[snet]->n_net));
				rtdbp->cd_cost = chgctab[snet];
				break;

		  case T_FIX:
		  case T_FLT:	rtdbp->cd_addr.oc_word = chgatab[snet].oc_word;
				rtdbp->cd_cost = chgctab[snet];
				break;

		  default:	bughalt("bad network type code");
		  }
		}

	pkt->oc_len = (CHMSNET * sizeof(struct chdrtb));
	if (pkt->oc_len > CHMAX)
		bughalt("too many entries to fit in DRT ANS");
	iob->i_bxfr = (pkt->oc_len + sizeof(ochpkt));
	signal(gwpid, chopc, iob);
}
